#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/miscdevice.h>
#include <asm/xen/hypercall.h>
#include <asm/xen/page.h>

static noinline pte_t *find_pte(unsigned long vaddr) {
	/* pgd_offset_k() doesn't work from module */
	pgd_t *pgd = pgd_offset(current->mm, vaddr);
	p4d_t *p4d = p4d_offset(pgd, vaddr);
	pud_t *pud = pud_offset(p4d, vaddr);
	pmd_t *pmd = pmd_offset(pud, vaddr);
	pte_t *pte = pte_offset_kernel(pmd, vaddr);
	return pte;
}

static u64 *my_page;
static u64 *dummy_page;
static u64 pte_maddr;

static volatile unsigned int syncval;

static int run_test_1(void) {
	unsigned long irq_flags;
	int ret;
	int success_count;
	struct mmuext_op extop;
	int i;

	pr_warn("test1: entry\n");

	local_irq_save(irq_flags);
	pr_warn("test1: start sync\n");
	for (i=0; 1; i++) {
		unsigned int syncval_old = syncval;

		if (syncval_old == 10)
			break;
		if (i >= 0x10000000) {
			pr_warn("test1: sync failed\n");
			local_irq_restore(irq_flags);
			return -ETIMEDOUT;
		}
		if ((syncval_old & 1) == 1)
			syncval = syncval_old + 1;
	}
	pr_warn("test1: sync finished\n");


	extop = (struct mmuext_op) {
		.cmd = MMUEXT_PIN_L1_TABLE,
		.arg1.mfn = virt_to_mfn(my_page),
		.arg2.src_mfn = 0x12345678 // marker for delay patch
	};
	pr_warn("test1: about to do PIN_L1_TABLE (virt=0x%lx, mfn=0x%lx, pfn=0x%lx)\n",
			(unsigned long)my_page, (unsigned long)extop.arg1.mfn, virt_to_pfn(my_page));
	ret = HYPERVISOR_mmuext_op(&extop, 1, &success_count, DOMID_SELF);
	pr_warn("test1: did PIN_L1_TABLE\n");
	local_irq_restore(irq_flags);
	if (ret != 0) {
		pr_warn("test1: PIN_L1_TABLE failed: %d\n", ret);
		return -EFAULT;
	}

	pr_warn("test1: reached end\n");
	return -EXFULL;
}

static noinline int try_put_kernel_nofault(u64 *ptr, u64 val) {
	__put_kernel_nofault(ptr, &val, u64, Efault);
	return 0;
Efault:
	return -EFAULT;
}

static int run_test_2(void) {
	unsigned long irq_flags;
	struct mmu_update mmureq;
	u64 virt_to_machine_res;
	u64 virt_to_machine_res_page2;
	int i;
	int ret;
	int success_count;
	struct mmuext_op extop;

	virt_to_machine_res = virt_to_machine(my_page).maddr;
	virt_to_machine_res_page2 = virt_to_machine(dummy_page).maddr;

	local_irq_save(irq_flags);
	pagefault_disable();
	pr_warn("test2: start sync\n");
	for (i=0; 1; i++) {
		unsigned int syncval_old = syncval;

		if (syncval_old == 10)
			break;
		if (i >= 0x10000000) {
			pr_warn("test2: sync failed\n");
			local_irq_restore(irq_flags);
			pagefault_enable();
			return -ETIMEDOUT;
		}
		if ((syncval_old & 1) == 0)
			syncval = syncval_old + 1;
	}
	pr_warn("test2: sync finished\n");
	mdelay(1);
	pr_warn("test2: delay done\n");

	mmureq = (struct mmu_update) {
		.ptr = pte_maddr,
		.val = 0x3 | virt_to_machine_res
	};
	ret = HYPERVISOR_mmu_update(&mmureq, 1, &success_count, DOMID_SELF);
	if (ret != 0) {
		pr_warn("test2: mmu_update to RW failed: %d\n", ret);
		local_irq_restore(irq_flags);
		pagefault_enable();
		return -EFAULT;
	}
	((volatile u64 *)my_page)[0] = 0x1 | virt_to_machine_res_page2;
	mmureq = (struct mmu_update) {
		.ptr = pte_maddr,
		.val = 0x5 | virt_to_machine_res
	};
	ret = HYPERVISOR_mmu_update(&mmureq, 1, &success_count, DOMID_SELF);
	if (ret != 0) {
		pr_warn("test2: mmu_update to RO failed: %d\n", ret);
		local_irq_restore(irq_flags);
		pagefault_enable();
		return -EFAULT;
	}
	if (try_put_kernel_nofault(my_page+0, 0x1 | virt_to_machine_res_page2)) {
		pr_warn("test2: TLB entry immediately gone?\n");
		local_irq_restore(irq_flags);
		pagefault_enable();
		return -EFAULT;
	}
	/* poll for PTE writeback */
	for (i = 0; 1; i++) {
		u64 val = ((volatile u64 *)my_page)[0];

		if (val & 0x4)
			break;
		if (i >= 0x10000000) {
			pr_warn("test2: PTE never changed???\n");
			local_irq_restore(irq_flags);
			pagefault_enable();
			return -EFAULT;
		}
	}
	/* only do it with one PTE - if we do it more often, stuff will blow up so bad
	 * that I'd need a proper serial console to grab the xen log messages
	 */
	for (i=0; i<1; i++) {
		if (try_put_kernel_nofault(my_page+i, 0x7 | virt_to_machine_res_page2)) {
			pr_warn("test2: TLB entry gone when validated PTE got written back :(\n");
			local_irq_restore(irq_flags);
			pagefault_enable();
			return -EFAULT;
		}
	}
	pr_warn("test2: stale TLB write worked :)\n");
	local_irq_restore(irq_flags);
	pagefault_enable();

	pr_warn("test2: sleeping for 10s (in case the next step crashes)...\n");
	msleep(10000);
	pr_warn("test2: trying to trigger debug message in Xen\n");


	extop = (struct mmuext_op) {
		.cmd = MMUEXT_UNPIN_TABLE,
		.arg1.mfn = virt_to_mfn(my_page)
	};
	pr_warn("test2: about to do UNPIN_TABLE\n");
	ret = HYPERVISOR_mmuext_op(&extop, 1, &success_count, DOMID_SELF);
	pr_warn("test2: did UNPIN_TABLE, returned %d\n", ret);

	return -EXFULL;
}

static long testdev_ioctl(struct file *file, unsigned int cmd, unsigned long arg) {
	switch (cmd) {
	case 0x13370001:
		return run_test_1();
	case 0x13370002:
		return run_test_2();
	default:
		return -EINVAL;
	}
}

static const struct file_operations testdev_fops = {
	.owner = THIS_MODULE,
	.unlocked_ioctl = testdev_ioctl
};

static struct miscdevice test_dev = {
	.name = "xen_cmpxchg_abac",
	.fops = &testdev_fops
};

static int init_test(void) {
	struct mmu_update mmureq;
	int ret;
	int success_count;
	u64 dummy_pte_maddr;

	my_page = (void*)__get_free_pages(GFP_KERNEL | __GFP_ZERO, 0);
	dummy_page = (void*)__get_free_pages(GFP_KERNEL, 0);
	if (!my_page || !dummy_page) {
		pr_warn("test: GFP failed???\n");
		return -ENOMEM;
	}
	pte_maddr = virt_to_machine(find_pte((unsigned long)my_page)).maddr;
	dummy_pte_maddr = virt_to_machine(find_pte((unsigned long)dummy_page)).maddr;

	mmureq = (struct mmu_update) {
		.ptr = pte_maddr,
		.val = 0x5 | virt_to_machine(my_page).maddr
	};
	ret = HYPERVISOR_mmu_update(&mmureq, 1, &success_count, DOMID_SELF);
	if (ret != 0) {
		pr_warn("test: mmu_update to RO failed: %d\n", ret);
		return -EFAULT;
	}
	mmureq = (struct mmu_update) {
		.ptr = dummy_pte_maddr,
		.val = 0x5 | virt_to_machine(dummy_page).maddr
	};
	ret = HYPERVISOR_mmu_update(&mmureq, 1, &success_count, DOMID_SELF);
	if (ret != 0) {
		pr_warn("test: second mmu_update to RO failed: %d\n", ret);
		return -EFAULT;
	}
	pr_warn("test: registering miscdev\n");
	return misc_register(&test_dev);
}

static void cleanup_test(void) {
	misc_deregister(&test_dev);
}

module_init(init_test);
module_exit(cleanup_test);
MODULE_LICENSE("GPL v2");
